const redis = require('redis');
const { promisifyAll } = require('bluebird');

promisifyAll(redis.RedisClient.prototype);
promisifyAll(redis.Multi.prototype);

module.exports = configuration => {
    const CHANNEL_KEY_UPDATE = '__keyevent@0__:set';

    const client = redis.createClient(configuration);
    const listener = redis.createClient(configuration);
    const databaseListeners = [];
    const channelListeners = [];
    const keyListeners = [];
    const subscribedChannels = [];
    const subscribedKeys = [];

    const setValue = (key, value, json = true) => client.setAsync(key, json ? JSON.stringify(value) : value);
    const getValue = async (key, json = true) => {
        try {
            const value = await client.getAsync(key);
            return json ? JSON.parse(value) : value;
        } catch (error) {
            return null;
        }
    };
    const mergeValue = async (key, value, json = true) => {
        const oldValue = await getValue(key);
        const mergedValue = Object.assign({}, oldValue, value);
        return setValue(key, mergedValue);
    };

    const publish = (channel, message) => client.publishAsync(channel, message);
    const alreadySubscribedChannel = channel => subscribedChannels.find(_channel => _channel === channel);
    const alreadySubscribedKey = key => subscribedKeys.find(_key => _key === key);

    const registerDatabaseListener = processMessage => {
        listener.subscribe(CHANNEL_KEY_UPDATE);
        databaseListeners.push({ channel: CHANNEL_KEY_UPDATE, processMessage });
    };

    const registerChannelListener = (channel, processMessage) => {
        if(!alreadySubscribedChannel(channel)) {
            subscribedChannels.push(channel);
            listener.subscribe(channel);
        }
        channelListeners.push({ channel, processMessage });
    };

    const registerKeyListener = (key, processMessage) => {
        if(!alreadySubscribedKey(key)) {
            subscribedKeys.push(key);
            listener.subscribe(CHANNEL_KEY_UPDATE, key);
        }
        keyListeners.push({ channel: CHANNEL_KEY_UPDATE, key, processMessage });
    };

    client.on('error', error => {
        console.error('redis error', error.message);
    });

    listener.on('message', async (channel, message) => {
        let registeredListeners = [];
        const messageKey = message;

        if(channel === CHANNEL_KEY_UPDATE) {
            registeredListeners = keyListeners.filter(keyListener => keyListener.key === messageKey);
            registeredListeners = registeredListeners.concat(databaseListeners);

            message = await getValue(messageKey);
        } else {
            registeredListeners = channelListeners.filter(listener => listener.channel === channel);
        }

        registeredListeners
            .forEach(({ channel, key, processMessage }) => processMessage(message, key || messageKey, channel));
    });

    return { setValue, getValue, mergeValue, publish, registerChannelListener, registerKeyListener, registerDatabaseListener };
};

